/*============================================================================*

	SUITE:		e-Scrabble
	FILE:		draw_utils.c
	
	COPYRIGHT NOTICE:
	
	Copyright (C) 2001 Players Inc, All Rights Reserved

	THIS MATERIAL IS PROVIDED FOR ACADEMIC USE ONLY, you may modify it
	provided this copyright notice is preserved, however you may not
	redistribute it, or use it for commercial purposes without prior
	written permission of the authors.

	Direct questions, comments, or bug reports to:
	aul@ecf.utoronto.ca
	ticku@ecf.utoronto.ca
	varia@ecf.utoronto.ca	

*============================================================================*/


#include <stdio.h>
#include <gtk/gtk.h>
#include "draw_utils.h"
#include "utils.h"
#include "pci_api.h"


static const short letter_values[26] = {
	1, 3, 3, 2, 1, 4, 2, 4, 1, 8, 5, 1, 3, 1, 1, 3, 10, 1, 1,
	1, 1, 4, 4, 8, 4, 10
};


static const int default_colors[6][3] = {
	{255, 0, 0},
	{0, 0, 255},
	{102, 204, 255},
	{255, 153, 204},
	{255, 102, 204},
	{204, 204, 153}
};

static const int default_tile_color[3] = {
	204, 153, 102
};

static const int default_new_tile_color[3] = {
	0, 0, 170
};


extern int rack_size;		/* size of rack (in tiles) */
extern Board_Square **board_square_state;		/* The board layout */
extern Coord tile_size;
extern Coord board_position;
extern Coord rack_position;
extern GtkWidget *window;
extern GtkWidget *space;
extern GdkPixmap *buffer;
extern Coord board_dim;
extern Tile *rack_state;						/* The local rack */
extern Location_Status **board_tile_state;		/* The tile layout */
extern Mover mover;
extern Bmove board_move;

static void draw_board_square (Board_Square *, int, int);
static void draw_rack ();
static void return_tile_to_rack (Tile);



/* procedure refresh_rack(t: pointer to Tile, num_t: int) returns void
 * requires: t initialized (t != NULL)
 * effects: returns given tile list to player's tray */

void
refresh_rack (Tile *t, int num_t)
{
	int i;	
	for( i=0; i<num_t; i++ )
	  {
	  	return_tile_to_rack(t[i]);
	  }
}


void remove_my_rack ()
{
	int i, j;
	for(i=0; i<board_dim.x; i++)
		for(j=0; j<board_dim.y; j++) {
			Location_Status *loc = &(board_tile_state[i][j]);
			if(loc->fixed == FALSE && loc->t != TILE_EMPTY) {
				loc->t = TILE_EMPTY;
				clear_tile(FALSE,i,j);
			}
		}
	for(i=0; i<rack_size; i++)
		if( rack_state[i] != TILE_EMPTY ) {
			rack_state[i] = TILE_EMPTY;
			clear_tile(TRUE,i,0);
		}
}


/* procedure draw_my_move(void) returns void
 * effects: updates gui with latest move
 * modifies: global board_tile_state ( fixes tiles on board in latest move
 *		and sets these tiles as new ), outputs graphics to screen
*/
		
void
draw_my_move ()
{
	int i, j;

	for( i=0; i<board_dim.x; i++ )
		for( j=0; j<board_dim.y; j++ )
			if( board_tile_state[i][j].new_m == TRUE ) {
				Location_Status *loc = &(board_tile_state[i][j]);
				int x_ret = board_position.x + i*tile_size.x;
				int y_ret = board_position.y + j*tile_size.y;

				loc->new_m = FALSE;
				draw_tile (buffer,loc,x_ret,y_ret,TILE_EMPTY);
  		
				gdk_draw_pixmap (space->window,
					space->style->fg_gc[GTK_WIDGET_STATE(space)],
					buffer, x_ret-1, y_ret-1,
					x_ret-1, y_ret-1,
					tile_size.x+2, tile_size.y+2);		
				
			}

	for ( i=0; i<board_move.tiles_in_move; i++ )
	  {
	  	int x = board_move.pos[i].x;
		int y = board_move.pos[i].y;
		int x_ret = board_position.x + x*tile_size.x;
		int y_ret = board_position.y + y*tile_size.y;
		Location_Status *loc = &board_tile_state[x][y];

	  	loc->fixed = TRUE;
	  	loc->new_m = TRUE;		

		draw_tile (buffer,loc,x_ret,y_ret,TILE_EMPTY);
  		
		gdk_draw_pixmap (space->window,
			space->style->fg_gc[GTK_WIDGET_STATE(space)],
			buffer, x_ret-1, y_ret-1,
			x_ret-1, y_ret-1,
			tile_size.x+2, tile_size.y+2);		
	  }
	
}

/* procedure total_refresh_board() returns void
 * effects: update board from GM and redraw entire board
 * modifies: global board_tile_state (updates tile characteristics),
 *		outputs graphics to screen */

void
total_refresh_board ()
{
	int i, j;
    	Tile **t;
	gboolean changed=FALSE;
	pci_get_tiles( &t );

	for( i=0; i<board_dim.x; i++ )
	  {
	 	for( j=0; j<board_dim.y; j++ )
		  {	
		    Location_Status *loc = &(board_tile_state[i][j]);
			loc->new_m = FALSE;
			if(loc->t != t[i][j]) {
				loc->t = t[i][j];
				changed = TRUE;
			}
			loc->fixed = ( t[i][j] == TILE_EMPTY )? FALSE : TRUE;
		  }
	  }


	for( i=0; i<board_dim.x; i++ )
	  {
		s_free(t[i]);
	  }
	s_free(t);

	if( changed == TRUE )
		redraw_board ();
}


void
remove_move( Bmove bm, gboolean mymove )
{
	int i;
	for( i=0; i<bm.tiles_in_move; i++ )
	  {
		Location_Status *loc = &(board_tile_state[bm.pos[i].x][bm.pos[i].y]);

		loc->new_m = FALSE;
		loc->fixed = FALSE;
		
		if ( mymove )
		  {
			return_tile_to_rack(loc->t);
		  }

		{
			int x_ret = board_position.x + bm.pos[i].x*tile_size.x;
			int y_ret = board_position.y + bm.pos[i].y*tile_size.y;
		  	draw_board_square (&board_square_state[bm.pos[i].x][bm.pos[i].y], x_ret, y_ret);		  

			gdk_draw_pixmap (space->window,
				space->style->fg_gc[GTK_WIDGET_STATE(space)],
				buffer, x_ret-1, y_ret-1,
				x_ret-1, y_ret-1,
				tile_size.x+2, tile_size.y+2);		

		}

		loc->t = TILE_EMPTY;
  		
	  }
}


/* procedure incremental_refresh_board(bm: Bmove) returns void
 * requires: bm properly initialized (pointers in struct != NULL)
 * effects: update board from GM and redraw tiles that have changed
 * modifies: global board_tile_state (updates tile characteristics),
 *		outputs graphics to screen */

void
incremental_refresh_board ( Bmove bm )
{
	int i, j;
	Tile **t;
	
	for( i=0; i<board_dim.x; i++ )
		for( j=0; j<board_dim.y; j++ )
			if( board_tile_state[i][j].new_m == TRUE ) {
				Location_Status *loc = &(board_tile_state[i][j]);
				int x_ret = board_position.x + i*tile_size.x;
				int y_ret = board_position.y + j*tile_size.y;

				loc->new_m = FALSE;
				draw_tile (buffer,loc,x_ret,y_ret,TILE_EMPTY);
  		
				gdk_draw_pixmap (space->window,
					space->style->fg_gc[GTK_WIDGET_STATE(space)],
					buffer, x_ret-1, y_ret-1,
					x_ret-1, y_ret-1,
					tile_size.x+2, tile_size.y+2);		
				
			}

	for( i=0; i<bm.tiles_in_move; i++ )
	  {
		Location_Status *loc = &(board_tile_state[bm.pos[i].x][bm.pos[i].y]);
		if ( loc->t != TILE_EMPTY && !loc->fixed )
		  {
			return_tile_to_rack(loc->t);
		  }
		loc->new_m = TRUE;
		loc->fixed = TRUE;
		loc->t = bm.t[i];
		
		{
		int x_ret = board_position.x + bm.pos[i].x*tile_size.x;
		int y_ret = board_position.y + bm.pos[i].y*tile_size.y;
		draw_tile (buffer,loc,x_ret,y_ret,TILE_EMPTY);
  		
		gdk_draw_pixmap (space->window,
			space->style->fg_gc[GTK_WIDGET_STATE(space)],
			buffer, x_ret-1, y_ret-1,
			x_ret-1, y_ret-1,
			tile_size.x+2, tile_size.y+2);		
		}
	  }
	
	
/*	pci_get_tiles( &t );
	for( i=0; i<board_dim.x; i++ )
	  {
	 	for( j=0; j<board_dim.y; j++ )
		  {
		    Location_Status *loc = &(board_tile_state[i][j]);
			if( loc->new_m )
				continue;
			if( loc->t == TILE_EMPTY && t[i][j] == TILE_EMPTY)
				continue;
			if( loc->t == t[i][j] && loc->fixed )
				continue;
			if( t[i][j] == TILE_EMPTY && !loc->fixed )
				continue;

			if ( (loc->t != TILE_EMPTY && !(loc->fixed)) &&
t[i][j] != TILE_EMPTY )
			  {
				return_tile_to_rack(loc->t);
				if ( loc->t == t[i][j] )
				  {
				  	loc->fixed = TRUE;
					continue;
				  }
			  }		


			if ( t[i][j] != TILE_EMPTY && loc->t != t[i][j] )
			  {
				int x_ret = board_position.x + i*tile_size.x;
				int y_ret = board_position.y + j*tile_size.y;
				loc->fixed = TRUE;
				loc->t = t[i][j];

				draw_tile (buffer,loc,x_ret,y_ret,TILE_EMPTY);
  		
				gdk_draw_pixmap (space->window,
					space->style->fg_gc[GTK_WIDGET_STATE(space)],
					buffer, x_ret-1, y_ret-1,
					x_ret-1, y_ret-1,
					tile_size.x+2, tile_size.y+2);	
				continue;
			  }
			
			{
			int x_ret = board_position.x + i*tile_size.x;
			int y_ret = board_position.y + j*tile_size.y;
		  	draw_board_square (&board_square_state[i][j], x_ret, y_ret);

			gdk_draw_pixmap (space->window,
				space->style->fg_gc[GTK_WIDGET_STATE(space)],
				buffer, x_ret-1, y_ret-1,
				x_ret-1, y_ret-1,
				tile_size.x+2, tile_size.y+2);									
			}
				
		  }
	  }



	for( i=0; i<board_dim.x; i++ )
	  {
		s_free(t[i]);
	  }
	s_free(t);
*/
}
 

/* procedure clear_tile(on_rack: gboolean, x:int, y:int) returns void
 * requires: 0 <= x < board_dim.x, 0 <= y < board_dim.y
 * modifies: outputs graphics to screen
 * effects: clears tile, uncovering blank rack or blank board underneath */

void
clear_tile (gboolean on_rack, int x, int y)
{
	int xx, yy;

	if (on_rack)
	  {
	  	xx = rack_position.x + x*tile_size.x;
		yy = rack_position.y;
		
		gdk_draw_rectangle (buffer,
					window->style->white_gc,
					TRUE,
					xx,
					yy,
					tile_size.x,
					tile_size.y);
		gdk_draw_rectangle (buffer,
					window->style->black_gc,
					FALSE,
					xx,
					yy,
					tile_size.x,
					tile_size.y);	  
	  }
	else
	  {
	  	xx = board_position.x + tile_size.x*x;
		yy = board_position.y + tile_size.y*y;
		
	  	draw_board_square (&board_square_state[x][y], xx, yy);
	  }

	gdk_draw_pixmap (space->window,
			space->style->fg_gc[GTK_WIDGET_STATE(space)],
			buffer, xx-1, yy-1,
			xx-1, yy-1,
			tile_size.x+2, tile_size.y+2);
		  
}


/* procedure redraw_board(void) returns void
 * effects: redraws playing area
 * modifies: outputs graphics to screen */

void
redraw_board ()
{
	int i, j;
	gdk_window_clear (space->window);

	gdk_draw_rectangle (buffer,
				space->style->black_gc,
				TRUE,
				0,
				0,
				space->allocation.width,
				space->allocation.height);

	draw_rack ();

	/* Loop Through All Grid Locations and Draw Board */
	for (i=0; i<board_dim.x; i++)
		for (j=0; j<board_dim.y; j++)
		  { 
			if (board_tile_state[i][j].t == TILE_EMPTY)
			  {
				draw_board_square (&(board_square_state[i][j]),
							board_position.x + (i*tile_size.x),
							board_position.y + (j*tile_size.y));
			  }
			else
				draw_tile (buffer, &(board_tile_state[i][j]),
							board_position.x + (i*tile_size.x),
							board_position.y + (j*tile_size.y), TILE_EMPTY);
		  }	

	/* Loop Through All Player Tiles and Draw Them */
	for (i=0; i<rack_size; i++)
	  {
		if (rack_state[i] == TILE_EMPTY ) {}
		else
		  {
			draw_tile (buffer, NULL,
						rack_position.x + (i*tile_size.x),
						rack_position.y, rack_state[i]);
		  }
	  }

	gdk_draw_pixmap (space->window,
		space->style->fg_gc[GTK_WIDGET_STATE(space)],
		buffer, 0, 0, 0, 0,
		space->allocation.width, space->allocation.height);


}


/* procedure draw_tile(target: pointer to GdkPixmap, loc:pointer to 
 *		Location_Status, x:int, y:int, t: Tile) returns void
 * requires: 0 <= x < board_dim.x, 0 <= y < board_dim.y, target != NULL
 * effects: draws tile to target buffer */

void
draw_tile (GdkPixmap *target, Location_Status* loc, int x, int y, Tile t)
{
	GdkGC *gc;
	GdkColor *color;
	GdkFont *gf;
	char font[50];
	char letter_val[3];
	Tile tdraw = TILE_EMPTY;
	if( loc != NULL )
		t = loc->t;

	if( target == mover.pixmap )
	  {
		gc = gdk_gc_new (mover.pixmap);
		gdk_window_set_back_pixmap(mover.window, mover.pixmap, 0);
	  }
	else
	  {
		gc = gdk_gc_new (space->window);
	  }
	  	
	color = (GdkColor *)s_malloc (sizeof (GdkColor));

	if (loc==NULL || !(loc->new_m))
	  {
		color->red = default_tile_color[0] * (65535/255);
		color->green = default_tile_color[1] * (65535/255);
		color->blue = default_tile_color[2] * (65535/255);
	  }
	else
	  {
		color->red = default_new_tile_color[0] * (65535/255);
		color->green = default_new_tile_color[1] * (65535/255);
		color->blue = default_new_tile_color[2] * (65535/255);
	  }

	color->pixel = (gulong)((color->red)*65536 + (color->green)*256
							+ color->blue);



	gdk_color_alloc (gtk_widget_get_colormap(space), color);
	gdk_gc_set_foreground (gc, color);


	gdk_draw_rectangle (target,
		gc,
		TRUE,
		x,
		y,
		tile_size.x,
		tile_size.y);

	gdk_draw_rectangle (target,
		window->style->black_gc,
		FALSE,
		x,
		y,
		tile_size.x,
		tile_size.y);


	gdk_gc_destroy (gc);


	/* Hardcoded font setup */
	if ((t & BLANK_MASK) != 0)
	  {

	  	tdraw = t & ~BLANK_MASK;
		gc = gdk_gc_new (space->window);
		color->red = 153 * (65535/255);
		color->green = 153 * (65535/255);
		color->blue = 153 * (65535/255);
		color->pixel = (gulong)(color->red*65536 + color->green*256
							+ color->blue);

		gdk_color_alloc (gtk_widget_get_colormap (space), color);
		gdk_gc_set_foreground (gc, color);
	  }
	else if ( loc==NULL || !(loc->new_m) )
	  {
	  	tdraw = t;
	  	gc = space->style->black_gc;
	  }
	else
	  {
	  	tdraw = t;
	  	gc = space->style->white_gc;
	  }

	if( tdraw != TILE_EMPTY )
	  {
		sprintf(font, "-*-helvetica-bold-r-*-*-%i-*-*-*-*-*-*-*", (int)(0.8*tile_size.x));
		gf = gdk_font_load(font);
		gdk_draw_text (target,
					gf,
					gc,
					x+0.15*tile_size.x,
					y+0.85*tile_size.y,
					&tdraw,
					1);
		gdk_font_unref (gf);

		if((t & BLANK_MASK) == 0)
		  {
			sprintf(letter_val, "%i", letter_values[(int)t-(int)('A')]);
			sprintf(font, "-*-helvetica-medium-r-*-*-%i-*-*-*-*-*-*-*", (int)(0.3*tile_size.x));
			gf = gdk_font_load(font);
			gdk_draw_text (target,
					gf,
					gc,
					x+(0.9-0.1*strlen(letter_val))*tile_size.x,
					y+0.95*tile_size.y,
					letter_val,
					strlen(letter_val));
			gdk_font_unref (gf);
		  }
	  }

	s_free (color);
	if ((t & BLANK_MASK) != 0)
		gdk_gc_destroy (gc);

		
}

/* procedure draw_rack(void) returns void
 * effects: draws rack
 * modifies: outputs graphics to screen */

static void
draw_rack ()
{
	int i;
	
	gdk_draw_rectangle (buffer,
				space->style->white_gc,
				TRUE,
				rack_position.x,
				rack_position.y,
				rack_size * tile_size.x,
				tile_size.y);
				
	for( i=0; i<rack_size; i++ )
	  {
		gdk_draw_rectangle (buffer,
					space->style->black_gc,
					FALSE,
					rack_position.x+i*tile_size.x,
					rack_position.y,
					rack_size * tile_size.x,
					tile_size.y);
	  }
}


/* procedure draw_board_square(bs: pointer to Board_Square, x:int, y:int) 	
 *		returns void
 * requires: bs to point to non-NULL Board_Square
 * effects: draws square on board
 * modifies: outputs graphics to screen */

static void
draw_board_square (Board_Square *bs, int x, int y)
{
	GdkGC *gc;
	GdkColor *color;
	GdkFont *gf;
	char font[50];
	int font_width, font_height;

	gc = gdk_gc_new (space->window);
	color = (GdkColor *)s_malloc (sizeof (GdkColor));


	color = (GdkColor *)s_malloc (sizeof (GdkColor));

	color->red = default_colors[bs->color][0] * (65535/255);
	color->green = default_colors[bs->color][1] * (65535/255);
	color->blue = default_colors[bs->color][2] * (65535/255);

	color->pixel = (gulong)(color->red*65536 + color->green*256
							+ color->blue);
	gdk_color_alloc (gtk_widget_get_colormap (space), color);
	gdk_gc_set_foreground (gc, color);

	gdk_draw_rectangle (buffer,
					gc,
					TRUE,
					x,
					y,
					tile_size.x,
					tile_size.y);
	gdk_draw_rectangle (buffer,
					window->style->black_gc,
					FALSE,
					x,
					y,
					tile_size.x,
					tile_size.y);

	if (strlen(bs->label) > 0)
	  {
		sprintf(font, "-*-helvetica-medium-r-*-*-%i-*-*-*-*-*-*-*", (int)(0.4*tile_size.x));
		gf = gdk_font_load(font);
		font_width = gdk_text_width (gf, bs->label, 3);
		font_height = gdk_text_height (gf, bs->label, 3);
		gdk_draw_text (buffer,
					gf,
					space->style->black_gc,
					x+(tile_size.x-font_width)/2,
					y+(tile_size.y+font_height)/2,
					bs->label,
					3);
		gdk_font_unref (gf);
	  }
	
	s_free (color);
	gdk_gc_destroy (gc);

}


/* procedure return_tile_to_rack(t: Tile) returns void
 * effects: finds place for tile on rack, and renders it there
 * modifies: global rack_state (updates status of rack),
 * 		outputs graphics to screen */

static void
return_tile_to_rack (Tile t)
{
	int j;

	for( j=0; j<rack_size; j++ )
	  {
	  	if( rack_state[j] == TILE_EMPTY )
		  {
		  	int x_ret = rack_position.x + j*tile_size.x;
			int y_ret = rack_position.y;

		  	rack_state[j] = t;
			draw_tile (buffer,NULL,x_ret,y_ret,t);
  		
			gdk_draw_pixmap (space->window,
				space->style->fg_gc[GTK_WIDGET_STATE(space)],
				buffer, x_ret-1, y_ret-1,
				x_ret-1, y_ret-1,
				tile_size.x+2, tile_size.y+2);
		  	break;
		  }
	  }
}

