/*============================================================================*
*============================================================================*


	SUITE:		e-Scrabble
	FILE:		pci_api.c

	COPYRIGHT NOTICE:

	Copyright (C) 2001 Players Inc, All Rights Reserved

	THIS MATERIAL IS PROVIDED FOR ACADEMIC USE ONLY, you may modify it
	provided this copyright notice is preserved, however you may not
	redistribute it, or use it for commercial purposes without prior
	written permission of the authors.

	Direct questions, comments, or bug reports to:
	aul@ecf.utoronto.ca
	ticku@ecf.utoronto.ca
	varia@ecf.utoronto.ca

*============================================================================*/

#include <stdio.h>
#include <ctype.h>

#include "pci_api.h"
#include "draw_utils.h"
#include "utils.h"
#include "internal.h"
#include <strings.h>

extern gboolean challenge_status;
extern gboolean make_move_status;
extern int rack_size;
extern Coord board_dim;
PlayerID myid = -1;

/*static void incoming_message (gpointer, gint, GdkInputCondition);*/



typedef struct _passstruct
 {
  int num_tiles;
  Tile letters[7];
 } PassStruct; 

char *status_messages[] = {
	"> Your challenge was UNSUCCESSFUL\n",
	"> There is no error-1\n",
	"> There is no error-2\n",
	"> There is no error-3\n",
	"> You are not the current player\n",
	"> Miscellaneous error\n",
	"> Unsuccessfull challenge\n",
	"> Invalid Move\n",
	"> The game has not started\n",
	"> The game has already started\n",
	"> You can't challenge now\n",
	"> Not enough players\n",
	"> The game has not started\n",
	"> The game has already started\n",
	"> Must have some tiles\n",
	"> Not enough tiles to swap\n",
	"> There was a problem saving the game\n",
	"> There was a problem restoring the game\n",
};



int                 sd,port;
int                 gm_tempsd,gm_sd,gm_port;
int                 answer;
char                *host;
struct hostent      *hp;
struct sockaddr_in  server;
struct sockaddr_in  gmserver;
char                *pname;
GeneralPacket       *send_packet;
GeneralPacket       *get_packet;
GeneralPacket       *answer_packet;

int                 packet_size;


/*---------------------------------------------------*/

int
check_from_blank(int a)
 {
  int b;

  b = a & 0x80;
  if (b == 0x80)
    return(1);
  else
    return(0);
 }

/*---------------------------------------------------*/

char
change_blank_encoding(int a)
 {
  return((char)(a & 0x7F));
 }

/*---------------------------------------------------*/

int
change_blank_encoding_inverse(char a)
 {
  
 
  return(a | 0x80);
 }

/*---------------------------------------------------*/


int
change_return_value(int a)
 {
  if (a >= OK)
    return(SUCCESS);
  else
    return(-a);
 }

/*---------------------------------------------------*/

int
pci_i_am_here (int port2, int gm_port2, char host2[])
 {
  int client_len;
  int on=1;
  struct sockaddr_in client;

  /* Initialize some variables */
  port = port2;
  host = host2;
  gm_port = gm_port2;
  packet_size = sizeof(struct packet);
  send_packet = (struct packet *)malloc(packet_size);
  get_packet  = (struct packet *)malloc(packet_size);
  answer_packet  = (struct packet *)malloc(packet_size);
 

  /* Create a socket */   
  if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
   {
    perror("Can't create socket\n");
    exit(1);
   }

  /* Get and store server's info */
  bzero((char *)&server, sizeof(struct sockaddr_in));
  server.sin_family = AF_INET;
  server.sin_port = htons(port);

  if ((hp = gethostbyname(host)) == NULL)
   {
    perror("Can't get server's address\n");
    exit(1);
   }

  bcopy((char *)hp->h_addr,(char *)&server.sin_addr.s_addr,hp->h_length);

  /* Connect to server */
  if (connect(sd, (struct sockaddr *)&server, sizeof(server)) == -1)
   {
    perror("Can't connect\n");
    exit(1);
   }
  
  /* Create a socket to accept connections on */
  if ((gm_tempsd = socket(AF_INET,SOCK_STREAM, 0)) < 0)
   {
    perror("Can't create a socket\n");
    exit(1);
   }

  /* Initialize server info */
  bzero((char *)&gmserver,sizeof(struct sockaddr_in));
  gmserver.sin_family = AF_INET;

  gmserver.sin_port = htons(gm_port);  
  gmserver.sin_addr.s_addr = htonl(INADDR_ANY);

  setsockopt(gm_tempsd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)); 

  /* Bind the port to accept connections on */
  if (bind(gm_tempsd,(struct sockaddr *)&gmserver,sizeof(gmserver)) == -1)
   {
    perror("Can't bind\n");
    exit(1);
   }

  /* Listen to the port for up to MAXPLAYERS connections */
  listen(gm_tempsd,MAXPLAYERS);


  /* Accept a new connectiona and return the new socket */
  if ((gm_sd = accept(gm_tempsd, (struct sockaddr *)&client, &client_len)) == -1)
   {
    perror("Can't accept client\n");
    exit(1);
   }

  return(SUCCESS);
 }

/*---------------------------------------------------*/

int
pci_challenge_move ()
 {
  int retVal = 0;

  /* Send GM Challenge Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = CHALLENGE;
  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }
   
  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }
 
  answer = get_packet->type;
  
  retVal = change_return_value(answer);
  
  if (answer == OK)
   return(SUCCESS);
  else
   return (retVal);
}

/*---------------------------------------------------*/


int
pci_new_game ( gboolean restore, PlayerIDList *pidlist )
 {
  PlayerPacket *pp;
  int          i;

  /* Send GM NewGame Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = NEWGAME;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  /* Can't start game, so return error */
  if (answer < 0)
   {
    return(change_return_value(answer));
   }

  /* Game has started, get player info */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = PLAYERINFO;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  
  /* Check for error, should not happen */
  if (answer_packet->type != PLAYERINFORETURN)
   {

    return(change_return_value(MISCERROR));
   }

  pp = (PlayerPacket *)answer_packet;
  pidlist = (PlayerIDList *)malloc(sizeof(PlayerIDList));
  pidlist->num_players = pp->numberOfPlayers;
  pidlist->players = (PlayerID *)malloc(pp->numberOfPlayers * sizeof(PlayerID));

  for (i = 0; i < pp->numberOfPlayers; i++)
   {
    pidlist->players[i] = pp->pid[i];
   }



  return(SUCCESS);
 }

/*---------------------------------------------------*/

int
pci_try_move (Bmove move, int *score)
 {
  BMovePacket *mp;
  ScoreReturnPacket *sp;

  int i;
    
  if (move.tiles_in_move == 0)
	return(14);

  bzero(send_packet,sizeof(GeneralPacket));

  mp = (BMovePacket *)send_packet;

  mp->type = TRYMOVE;
  mp->theMove.type = PLACE;
  mp->theMove.numMoves = move.tiles_in_move;

  for (i = 0; i < move.tiles_in_move; i++)
   {
    if (check_from_blank(move.t[i]))
     {
      mp->theMove.moves[i].letter = tolower(change_blank_encoding(move.t[i]));
      mp->theMove.moves[i].fromBlank = 1;
     }
    else
     {
      mp->theMove.moves[i].letter = tolower(move.t[i]);
      mp->theMove.moves[i].fromBlank = 0;
     }

    mp->theMove.moves[i].Pos.x = move.pos[i].x;
    mp->theMove.moves[i].Pos.y = move.pos[i].y;
   }

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }
   
   
  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  sp = (ScoreReturnPacket *)answer_packet;
  *score = sp->score;

  /* Check for error */
  if (*score < 0)
   {
    return(change_return_value(*score));
   }

  return(SUCCESS);
 }

/*---------------------------------------------------*/

int
pci_make_move (Bmove move, Scores *s)
 {
  BMovePacket       *mp;
  PlayerPacket      *pp;
  Tile t[7];
  int temp;

  int i;
  
  if (move.is_swap == TRUE)
   {
    for (i = 0 ; i <move.tiles_in_move; i++)
     {
      t[i] = tolower(move.t[i]);
     }
    temp = pci_send_tiles(t,move.tiles_in_move);  
	if (temp != SUCCESS)
	  return(temp);
   }
  else
   {
    
    if (move.tiles_in_move == 0)
     {
      pci_get_scores(s);
	 
      return(14);  /* don't know why this works */
     } 
   
    bzero(send_packet,sizeof(GeneralPacket));

    mp = (BMovePacket *)send_packet;

    mp->type = BMOVE;

    mp->theMove.type = PLACE;
    mp->theMove.numMoves = move.tiles_in_move;

    for (i = 0; i < move.tiles_in_move; i++)
     {

      if (check_from_blank(move.t[i]))
       {
        mp->theMove.moves[i].letter =(change_blank_encoding(move.t[i]));
        mp->theMove.moves[i].fromBlank = 1;
       }
      else
       {
        mp->theMove.moves[i].letter = tolower(move.t[i]);
        mp->theMove.moves[i].fromBlank = 0;
       }

      mp->theMove.moves[i].Pos.x = move.pos[i].x;
      mp->theMove.moves[i].Pos.y = move.pos[i].y;
     }


    /* Send the packet */
    if (write(sd,send_packet,packet_size) < 0)
     {
      perror("Write failed\n");
      exit(1);
     }

    if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
     {
      perror("Write failed\n");
      exit(1);
     }

    
    answer = get_packet->type;

    bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

    /* Check for error */
    if (answer_packet->type < 0)
     {

      return(change_return_value(answer_packet->type));
     }
   }


  /* Get scores */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = PLAYERINFO;


  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }


  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));


  /* Check for error, should not happen */
  if (answer_packet->type != PLAYERINFORETURN)
   {

    return(change_return_value(MISCERROR));
   }

  pp = (PlayerPacket *)answer_packet;
  s->num_players = pp->numberOfPlayers;
  s->players = (PlayerID *)malloc(pp->numberOfPlayers * sizeof(PlayerID));
  s->scores  = (int *)malloc(pp->numberOfPlayers * sizeof(int));


  for (i = 0; i < pp->numberOfPlayers; i++)
   {
    s->players[i] = pp->pid[i];
    s->scores[i]  = pp->score[i];
   }

  return(SUCCESS);
 }

/*---------------------------------------------------*/

int
pci_send_tiles (Tile *t, int num_tiles)
 {
  SMovePacket *sp;
  int i;
  int retVal;

  bzero(send_packet,sizeof(GeneralPacket));

  sp = (SMovePacket *)send_packet;

  sp->type = SMOVE;
  sp->theMove.type = SWAP;
  sp->theMove.numMoves = num_tiles;

  for (i = 0; i < num_tiles; i++)
   {
    if (check_from_blank(t[i]))
      sp->theMove.letters[i] = '?';
    else
      sp->theMove.letters[i] = tolower(t[i]);
   }

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  /* Get back SwapMove status in retVal */

  if (answer == OK)
    retVal = SUCCESS;
  else	
    retVal = change_return_value(answer);
  return (retVal);

 }

/*---------------------------------------------------*/

int
pci_create_AP()
 {
  int retVal;

  bzero(send_packet,sizeof(GeneralPacket));

  send_packet->type = CREATEAUTOMATEDPLAYER;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;

  /* Get back Resign status in retVal */
  retVal = change_return_value(answer);

  if (answer == CREATEAUTOMATEDPLAYERRETURN)
   {
    return(SUCCESS);
   }	
  else
    return (retVal);

 }
 
/*---------------------------------------------------*/

int
pci_cancel_game ()
 {
  int retVal;

  bzero(send_packet,sizeof(GeneralPacket));

  send_packet->type = RESIGN;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;

  /* Get back Resign status in retVal */
  retVal = change_return_value(answer);

  if (answer == OK)
   {
    myid = -2;
    return(SUCCESS);
   }	
  else
    return (retVal);

 }
 

/*---------------------------------------------------*/

int
pci_start_game ( Scores *s )
 {

  /* Send GM NewGame Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = NEWGAME;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  /* Can't start game, so return error */
  if (answer < 0)
   {
    return(change_return_value(answer));
   }

  pci_get_scores(s);

  return(SUCCESS);

}

/*---------------------------------------------------*/

int
pci_save_game ()
 {
  /* Send GM NewGame Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = SAVEGAME;


  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if (read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;

  /* Can't start game, so return error */
  if (answer < 0)
   {
    return(change_return_value(answer));
   }

  return(SUCCESS);

 }

/*---------------------------------------------------*/

int
pci_restore_game (Scores *s)
 {
  /* Send GM NewGame Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = RESTOREGAME;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if (read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  
  /* Can't start game, so return error */
  if (answer < 0)
   {
    return(change_return_value(answer));
   }

  pci_get_scores(s);
  myid = -2;
  return(SUCCESS);

 }

/*---------------------------------------------------*/

int
pci_invite_player ( PlayerID pid )
 {
  return(0);
 }

/*---------------------------------------------------*/

int
pci_accept_invite (gboolean accept)
 {
  return(0);
 }

/*---------------------------------------------------*/  

int
pci_get_tiles (Tile ***t)
 {
  BoardPacket *bp;
  int i;
  int j;
  

  (*t) = s_malloc( board_dim.x * sizeof(Tile *) );
  for( i=0; i<board_dim.x; i++ )
   {
    (*t)[i] = s_malloc( board_dim.y * sizeof(Tile) );
   }


  /* Send GM SHOWBOARD Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = SHOWBOARD;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  /* Check for error, should not happen */
  if (answer_packet->type !=SHOWBOARDRETURN)
   {

    return(change_return_value(MISCERROR));
   }

  bp = (BoardPacket *)answer_packet;

  /* Copy the board */
  for (i = 0; i < LENGTH; i++) 
   for (j = 0; j < LENGTH; j++)
     (*t)[i][j] =toupper(bp->board[i][j]);

  return(SUCCESS);
}

/*---------------------------------------------------*/

int
pci_get_scores (Scores *s)
 {
  PlayerPacket *pp;
  int i;

  /* Send GM SCOREINFO Request */
  bzero(send_packet,sizeof(GeneralPacket));
  send_packet->type = SCOREINFO;

  /* Send the packet */
  if (write(sd,send_packet,packet_size) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

   if ( read(sd,get_packet,sizeof(GeneralPacket)) < 0)
   {
    perror("Write failed\n");
    exit(1);
   }

  answer = get_packet->type;
  bcopy(get_packet,answer_packet,sizeof(GeneralPacket));

  /* Check for error, should not happen */
  if (answer_packet->type < 0)
   {
    return(change_return_value(MISCERROR));
   }

  pp = (PlayerPacket *)answer_packet;
  s->num_players = pp->numberOfPlayers;
  s->players = (PlayerID *)malloc(sizeof(PlayerID) * pp->numberOfPlayers);
  s->scores = (int *)malloc(sizeof(int) * pp->numberOfPlayers);

  for (i = 0; i < pp->numberOfPlayers; i++)
   {
    s->players[i] = pp->pid[i];
    s->scores[i]  = pp->score[i];
   }

  return(SUCCESS);

 }

/*---------------------------------------------------*/

void
gm_request_move ()
 {

  make_move_status = TRUE;
  print_msg( "> Please make a move\n" );

 }
 
/*---------------------------------------------------*/

void
gm_invite_challenge ()
 {
  print_msg ( "> You have been invited to challenge\n" );
  challenge_status = TRUE;
 }

/*---------------------------------------------------*/

void
gm_game_over (WinnerPacket *packet)
 {
  char pid_str[10];
  Scores s;
  int winner;
  int i;
  
  winner = packet->winner;
	

  s.num_players = packet->numberOfPlayers;
  s.players = (PlayerID *)malloc(sizeof(PlayerID) * packet->numberOfPlayers);
  s.scores = (int *)malloc(sizeof(int) * packet->numberOfPlayers);

  for (i = 0; i < packet->numberOfPlayers; i++)
   {
    s.players[i] = packet->pid[i];
    s.scores[i]  = packet->score[i];
   }  
  
  print_scores(&s);
  
  if (winner == myid)
  	print_msg( "\n> GAME OVER!!\n> You are the winner of the game\n" );  
  else if (winner > 0)
   {
    print_msg( "\n> GAME OVER!!\n> Winner of Game: " );
    sprintf (pid_str, "%i\n", winner);
    print_msg( pid_str );
   }
  else
   print_msg( "\n> GAME OVER!!\n> Automatic Player won the game\n" ); 
   
  myid = -2;
 }
 
 


/*--------------------------------------------------*/

void
gm_user_invited ( PlayerID invitor )
 {
  /* THE GM WILL NEVER CALL THIS FUNCTION */
  get_invitation_response( invitor );
 }

/*--------------------------------------------------*/

void
gm_move_revoked (Bmove mv, PlayerID pid)
 {
  Scores s;
  gboolean mymove;

  print_msg( "\n> A move has been revoked\n" );
  
  mymove = (pid == myid)?TRUE:FALSE;
  remove_move (mv, mymove);

  if (pci_get_scores(&s) == SUCCESS)
    print_scores(&s);
 }
 

 

/*--------------------------------------------------*/

void
gm_player_update ()
 {
  Scores s;
  
  print_msg( "\n\n>Players have been updated\n" );

  if (pci_get_scores(&s) == SUCCESS)
    print_scores(&s);

 }



/*--------------------------------------------------*/

void
gm_new_move (Bmove move)
 {
  Scores s;
  
  print_msg( "> A new move has been played\n\n" );

  incremental_refresh_board ( move );

  if (pci_get_scores(&s) == SUCCESS)
    print_scores(&s);

 }
 

 

/*--------------------------------------------------*/

void
gm_start_game (gboolean start, PlayerID pid)
 {
  Scores s;
  
  if (start)
   {
    if( myid != pid )
     {
      char msg[80];
      total_refresh_board (); 
	  remove_my_rack();
      myid = pid;
      print_msg( "> Lets Play e-Scrabble!!\n" );
      sprintf(msg, "> Your Player ID is %i\n\n", pid);
      print_msg( msg );
     }
	else
	 {
	  total_refresh_board();
	  remove_my_rack();
	 }
	
    if (pci_get_scores(&s) == SUCCESS)
      print_scores(&s);
   }
  else
   {
    print_msg( "> Game was cancelled\n" );
   }
 }



/*--------------------------------------------------*/
/* Will never get called in this GM ----------------*/
/*--------------------------------------------------*/
void
gm_request_tiles ()
 {
  int num;
  Tile *t = s_malloc( rack_size * sizeof(Tile) );
  get_rack_tiles( t, &num ) ;
  pci_send_tiles( t, num );
  s_free( t );
 }

/*--------------------------------------------------*/


void
gm_refresh_tray ( Tile *t, int num_tiles )
 {
  refresh_rack(t, num_tiles);	
 }




/*--------------------------------------------------*/


void
setup_listener ()
 {
  int source=gm_sd;



  gdk_input_add ( source, GDK_INPUT_READ,incoming_message, NULL );
}


/*--------------------------------------------------*/

void
incoming_message(gpointer data, gint source, GdkInputCondition cond)
 {
  ChallengePacket *cp;
  MoveCompletedPacket *mcp;
  TrayPacket *tp=(TrayPacket *)s_malloc(sizeof(GeneralPacket));
  Bmove temp;
  int i;

   
  if ( read(gm_sd,get_packet,sizeof(GeneralPacket)) < 0)
  	return;
   
  if (get_packet->type == YOURMOVE)
	gm_request_move();
  else if (get_packet->type == GAMENOWSTARTED)
   {

    gm_start_game(1,((GameStartedPacket *)get_packet)->pid);
   }
  else if (get_packet->type == CHALLENGEREQUEST)
    gm_invite_challenge();
  else if (get_packet->type == WINNER)
    gm_game_over(((WinnerPacket *)get_packet));
  else if (get_packet->type == PLAYERHASRESIGNED)
    gm_player_update();
  else if (get_packet->type == SUCCESSFULLCHALLENGE)
   {
    cp = (ChallengePacket *)get_packet;

    temp.is_swap = 0;
    temp.pos = (Coord *)malloc(sizeof(Coord) * cp->theMove.numMoves);
    temp.t   = (char *)malloc(sizeof(char) * cp->theMove.numMoves);
    temp.tiles_in_move = cp->theMove.numMoves;

    for (i = 0; i < temp.tiles_in_move; i++)
     {
      temp.pos[i].x = cp->theMove.moves[i].Pos.x;
      temp.pos[i].y = cp->theMove.moves[i].Pos.y;
		
      cp->theMove.moves[i].letter = toupper(cp->theMove.moves[i].letter);

      if (cp->theMove.moves[i].fromBlank)
        temp.t[i]   = change_blank_encoding_inverse(cp->theMove.moves[i].letter);
      else
        temp.t[i]   = cp->theMove.moves[i].letter;
     }
	   
    gm_move_revoked(temp, cp->playerID);
   }
  else if (get_packet->type == BMOVECOMPLETED)
   {
    mcp = (MoveCompletedPacket *)get_packet;

    temp.is_swap = 0;
    temp.pos = (Coord *)malloc(sizeof(Coord) * mcp->theMove.numMoves);
    temp.t   = (char *)malloc(sizeof(char) * mcp->theMove.numMoves);
    temp.tiles_in_move = mcp->theMove.numMoves;

    for (i = 0; i < temp.tiles_in_move; i++)
     {
      temp.pos[i].x = mcp->theMove.moves[i].Pos.x;
      temp.pos[i].y = mcp->theMove.moves[i].Pos.y;
		
      mcp->theMove.moves[i].letter = toupper(mcp->theMove.moves[i].letter);

      if (mcp->theMove.moves[i].fromBlank)
        temp.t[i]   = change_blank_encoding_inverse(mcp->theMove.moves[i].letter);
      else
        temp.t[i]   = mcp->theMove.moves[i].letter;
     }

    gm_new_move(temp);

   }
  else if (get_packet->type == SMOVECOMPLETED)
   {
    temp.is_swap = 1;
    temp.tiles_in_move = 0;
    gm_new_move(temp);
   }
  else if (get_packet->type == NEWTILES)
   {
    int i;

    bcopy(get_packet,tp,sizeof(GeneralPacket));
	 
    for (i = 0; i < tp->numberOfTiles; i++)
     {
      if ((tp->letters[i] == '?') || (tp->letters[i] == 0))
        tp->letters[i] = 0x80;
      else
        tp->letters[i] = toupper(tp->letters[i]);
     }	  
		
    gm_refresh_tray(tp->letters,tp->numberOfTiles);
   }

  return;
 }

