/*============================================================================*

	SUITE:		e-Scrabble
	FILE:		scrabble.c
	
	COPYRIGHT NOTICE:
	
	Copyright (C) 2001 Players Inc, All Rights Reserved

	THIS MATERIAL IS PROVIDED FOR ACADEMIC USE ONLY, you may modify it
	provided this copyright notice is preserved, however you may not
	redistribute it, or use it for commercial purposes without prior
	written permission of the authors.

	Direct questions, comments, or bug reports to:
	aul@ecf.utoronto.ca
	ticku@ecf.utoronto.ca
	varia@ecf.utoronto.ca	

*============================================================================*/



/*========================= INCLUDES =========================*/

#include <stdio.h>
#include "scrabble.h"
#include "utils.h"
#include "menu.h"
#include "user_events.h"
#include "draw_utils.h"
#include "pci_api.h"


/*========================= DEFINES ==========================*/

#define RELEASE		1
#define PRESS		2
#define UCASE(t)	((t) - 'a' + 'A');


/*======================== CONSTANTS =========================*/

static const int default_board_dim_x = 15;
static const int default_board_dim_y = 15;
static const int default_rack_size = 7;

static const float board_aspect_ratio = 0.90;
static const int default_border_width = 5;

static char* const board_text[] = {
	"TW",
	"TL",
	"DL",
	"DW",
	"",
	""
};

static const short default_board_stats[15][15] = {
	{0, 5, 5, 2, 5, 5, 5, 0, 5, 5, 5, 2, 5, 5, 0},
	{5, 3, 5, 5, 5, 1, 5, 5, 5, 1, 5, 5, 5, 3, 5},
	{5, 5, 3, 5, 5, 5, 2, 5, 2, 5, 5, 5, 3, 5, 5},
	{2, 5, 5, 3, 5, 5, 5, 2, 5, 5, 5, 3, 5, 5, 2},
	{5, 5, 5, 5, 3, 5, 5, 5, 5, 5, 3, 5, 5, 5, 5},
	{5, 1, 5, 5, 5, 1, 5, 5, 5, 1, 5, 5, 5, 1, 5},
	{5, 5, 2, 5, 5, 5, 2, 5, 2, 5, 5, 5, 2, 5, 5},
	{0, 5, 5, 2, 5, 5, 5, 4, 5, 5, 5, 2, 5, 5, 0},
	{5, 5, 2, 5, 5, 5, 2, 5, 2, 5, 5, 5, 2, 5, 5},
	{5, 1, 5, 5, 5, 1, 5, 5, 5, 1, 5, 5, 5, 1, 5},
	{5, 5, 5, 5, 3, 5, 5, 5, 5, 5, 3, 5, 5, 5, 5},
	{2, 5, 5, 3, 5, 5, 5, 2, 5, 5, 5, 3, 5, 5, 2},
	{5, 5, 3, 5, 5, 5, 2, 5, 2, 5, 5, 5, 3, 5, 5},
	{5, 3, 5, 5, 5, 1, 5, 5, 5, 1, 5, 5, 5, 3, 5},
	{0, 5, 5, 2, 5, 5, 5, 0, 5, 5, 5, 2, 5, 5, 0}	
};



/*======================== GLOBALS =========================*/

extern char *status_messages[];
extern PlayerID myid;

static int button_down = 0;
static GtkWidget *table;
static GtkWidget *t_Scores;
static GtkWidget *t_Messages;
static GtkWidget *text;
static GtkWidget *dialog;


gboolean challenge_status;
gboolean make_move_status;
int rack_size;		/* size of rack (in tiles) */
Coord board_dim;		/* dimensions of board (in tiles) */
Bmove board_move;
Board_Square **board_square_state;		/* The board layout */
Location_Status **board_tile_state;		/* The tile layout */
Tile *rack_state;						/* The local rack */
Coord tile_size;
Coord board_position;
Coord rack_position;
GtkWidget *window;
GtkWidget *space;
GtkWidget *vbox;
GdkPixmap *buffer;
Mover mover;



/*============== PRIVATE FUNCTION PROTOTYPES ==============*/

static void load_game (int, int, char[]);
static void setup_board ();
static void setup_rack ();
static void setup_move ();
static void free_mem ();
static void clear_board ();
static void set_board_squares ();
static void refresh (GtkWidget *);
static void swap_tile (int, int);
static gboolean setup_mover (int, int, int);
static gboolean get_tile_stats (int, int, int *, int *, int *, int *, int *, Tile *);
static gboolean valid_drop (int, int, int *, int *, int *, Tile);
static gboolean get_tile_location (int, int, int *, int *, int *);
static void get_blank ();
static void get_letter (GtkWidget *, gpointer);
static void create_window ();
static void create_space ();
static void create_mover ();
static void create_buffer (GtkWidget *);
static gint expose_space(GtkWidget *, GdkEventExpose *);
static gint configure_space(GtkWidget *, GdkEventConfigure *);
static gint button_press_space(GtkWidget *, GdkEventButton *);
static gint button_release_space(GtkWidget *, GdkEventButton *);
static gint motion_notify_space(GtkWidget *, GdkEventButton *);
static gboolean is_valid_tile (Tile *);
static void invitation_response ( GtkWidget *, gpointer );

/* -------------- FUNCTION DEFINITIONS -------------- */

int
main (int argc, char *argv[])
{
	buffer = NULL;
	challenge_status = FALSE;
	make_move_status = FALSE;

/*	if (argc != 4)
	 {
	  printf("Usage %s <message> <hostname> <server port> <client port>\n", argv[0]);
	  exit(1);
	 }*/

	if (argc != 2)
	 {
	  printf("Usage %s <message> <hostname>\n", argv[0]);
	  exit(1);
	 }


/*	load_game (atoi(argv[2]),atoi(argv[3]),argv[1]);*/
        load_game (5432,7890,argv[1]);
	setup_board ();
	setup_rack ();
	setup_move ();

	gtk_init ( &argc, &argv );	
	gdk_init ( &argc, &argv );

	setup_listener ();
	create_window ();
	create_menu ();
	create_space ();

	print_msg("Welcome to e-Scrabble\n\n");

	gtk_widget_show (window);

	create_mover ();
	
	gtk_main ();

	free_mem ();

	return (0);
}



static void
load_game (int sport, int myport, char host[])
{
	int retVal;

	retVal = pci_i_am_here (sport, myport,host);
	
	if ( retVal != SUCCESS )
	  {
		printf("%s\n", status_messages[retVal]);
		exit(-1);
	  }
}


void
get_rack_tiles( Tile *t, int *num_tiles )
{
	int i, j;
	(*num_tiles) = 0;
	
	for (i=0; i<board_dim.x; i++)
		for (j=0; j<board_dim.y; j++)
			if (board_tile_state[i][j].t != TILE_EMPTY &&
										!board_tile_state[i][j].fixed)
			  	t[(*num_tiles)++] = board_tile_state[i][j].t;
				
	for (i=0; i<rack_size; i++)
		if (rack_state[i] != TILE_EMPTY)
			t[(*num_tiles)++] = rack_state[i];
			
}


static void
create_mover ()
{
	GdkWindowAttr attr;
	
	attr.wclass = GDK_INPUT_OUTPUT;
	attr.window_type = GDK_WINDOW_CHILD;
	attr.event_mask = 0;
	attr.width = tile_size.x;
	attr.height = tile_size.y;
	attr.colormap = gdk_window_get_colormap (space->window);
	attr.visual = gdk_window_get_visual (space->window);

	mover.window = gdk_window_new (space->window, &attr,
							(GDK_WA_VISUAL | GDK_WA_COLORMAP));
	
	mover.t = TILE_EMPTY;	
}

static void
swap_tile (int x, int y)
{
	int on_rack, xx, yy;
	if( get_tile_location( x, y, &on_rack, &xx, &yy ) )
	  {
		if( on_rack )
		  {
			board_move.t[board_move.tiles_in_move++] = rack_state[xx];
		  	rack_state[xx] = TILE_EMPTY;	
			clear_tile (on_rack, xx, yy);
		  }
		else
		  {
		  	if( board_tile_state[xx][yy].t != TILE_EMPTY && !board_tile_state[xx][yy].fixed )
			  {
				board_move.t[board_move.tiles_in_move++] = board_tile_state[xx][yy].t;
				board_tile_state[xx][yy].t = TILE_EMPTY;
				clear_tile (on_rack, xx, yy);
			  }
		  }
	  }
}


static gboolean
get_tile_location (int x, int y, int *on_rack, int *xx, int *yy)
{
	int x_1, y_1;
	int x_2, y_2;
	
	x_1 = x - board_position.x;
	y_1 = y - board_position.y;
	
	x_2 = x - rack_position.x;
	y_2 = y - rack_position.y;
	
	if ( x_1 < 0 || y_1 < 0 ||
		x_1 > (board_dim.x*tile_size.x) || y_1 > (board_dim.y*tile_size.y))
	  {
	  	if ( x_2 < 0 || y_2 < 0 ||
			x_2 > (rack_size*tile_size.x) || y_2 > tile_size.y )
		  {
		  	return (FALSE);
		  }
		else
		  {
		  	*on_rack = TRUE;
			*xx = x_2 / tile_size.x;
		  }		  
	  }
	else
	  {
		*on_rack = FALSE;
		*xx = x_1 / tile_size.x;
		*yy = y_1 / tile_size.y;
	  }
	
	
	
	return (TRUE);	

}


static gboolean
get_tile_stats (int x, int y, int *on_rack, int *xx, int *yy,
					int *xoff, int *yoff, Tile *t)
{
	if ( get_tile_location( x, y, on_rack, xx, yy ) )
	  {
		if (*on_rack)
		  {
			int x_2 = x - rack_position.x;
			int y_2 = y - rack_position.y;

			if ( (*t = rack_state[*xx]) == TILE_EMPTY )
			  {
				return (FALSE);
			  }
			
			rack_state[*xx] = TILE_EMPTY;
			*xoff = x_2 % tile_size.x;
			*yoff = y_2;		  
		  }
		else
		  {
			int x_1 = x - board_position.x;
			int y_1 = y - board_position.y;

			if ( (*t = board_tile_state[*xx][*yy].t) == TILE_EMPTY ||
				board_tile_state[*xx][*yy].fixed || 
				board_tile_state[*xx][*yy].new_m)
			  {
				return (FALSE);
		  	  }
			  
			board_tile_state[*xx][*yy].t = TILE_EMPTY;
			*xoff = x_1 % tile_size.x;
			*yoff = y_1 % tile_size.y;
		  }
		return(TRUE);
	  }
	  
	return(FALSE);
}



static gboolean
valid_drop (int x, int y, int *on_rack, int *xx, int *yy, Tile t)
{	
	if (get_tile_location (x, y, on_rack, xx, yy))
	  {
	    if ( *on_rack )
		  {
			if ( rack_state[*xx] != TILE_EMPTY )
		  	  {
				return (FALSE);
		  	  }
			rack_state[*xx] = t;
		  }
		else
		  {
			if ( board_tile_state[*xx][*yy].t != TILE_EMPTY )
			  {
				return (FALSE);
		  	  }
			board_tile_state[*xx][*yy].t = t;
		  }
		return (TRUE);
	  }
	
	return (FALSE);	
}



static gboolean
setup_mover (int x, int y, int status)
{

	if (status == PRESS)
	  {
		if( !get_tile_stats (x, y, &mover.rackstart, &mover.xstart,
							&mover.ystart, &mover.xoff, &mover.yoff, &mover.t) )
		  {
		  	return (FALSE);
		  }
		else
		  {
			mover.pixmap = gdk_pixmap_new(mover.window, tile_size.x,
									tile_size.y,
									gdk_window_get_visual(mover.window)->depth);
		  	gdk_window_move (mover.window,x-mover.xoff,y-mover.yoff);
			draw_tile (mover.pixmap,NULL,0,0,mover.t);
			gdk_window_show(mover.window);
			clear_tile (mover.rackstart, mover.xstart, mover.ystart);
		  }		
	  }

	if (status == RELEASE)
	  {
		int x_ret, y_ret;
		gboolean is_blank = FALSE;
		if( valid_drop(x, y, &mover.racktarget, &mover.xtarget,
						&mover.ytarget,  mover.t) )
		  {
			if( mover.racktarget )
			  {
			    x_ret = rack_position.x + mover.xtarget*tile_size.x;
				y_ret = rack_position.y;
			  }
			else
			  {
			  	if( (mover.t & BLANK_MASK) != 0 )
				  {
				  	is_blank = TRUE;
				  }
			  	x_ret = board_position.x + mover.xtarget*tile_size.x;
				y_ret = board_position.y + mover.ytarget*tile_size.y;
			  }
		  }
		else
		  {
		    if( mover.rackstart )
			  {
			  	x_ret = rack_position.x + mover.xstart*tile_size.x;
				y_ret = rack_position.y;
				rack_state[mover.xstart] = mover.t;
			  }
			else
			  {
			  	x_ret = board_position.x + mover.xstart*tile_size.x;
				y_ret = board_position.y + mover.ystart*tile_size.y;
				board_tile_state[mover.xstart][mover.ystart].t = mover.t;
			  }
		  }

		if( is_blank )
		  {
			get_blank();
			if ( !is_valid_tile( &mover.t ) )
				mover.t = 'A' | BLANK_MASK;
			board_tile_state[mover.xtarget][mover.ytarget].t = mover.t;
		  }

		draw_tile (buffer,NULL,x_ret,y_ret,mover.t);
  		
		gdk_draw_pixmap (space->window,
				space->style->fg_gc[GTK_WIDGET_STATE(space)],
				buffer, x_ret-1, y_ret-1,
				x_ret-1, y_ret-1,
				tile_size.x+2, tile_size.y+2);

	  	gdk_window_hide(mover.window);
		if( mover.pixmap )
			gdk_pixmap_unref(mover.pixmap);
		mover.pixmap = NULL;

	  }
	 
	return (TRUE);
}


static gboolean
is_valid_tile ( Tile *t )
{
	gboolean is_blank;
	if ( ((*t) & BLANK_MASK) == 0 )
	  {
		is_blank = FALSE;
	  }
	else
	  {
		(*t) &= ~BLANK_MASK;
		is_blank = TRUE;
	  }
	  
	if ( (*t) >= 'a' && (*t) <= 'z' )
	  {
		*t = UCASE(*t);  
	  }
	else if ( (*t) >= 'A' && (*t) <= 'Z' ) {}
	else
	  {
		return FALSE;
	  }

	if (is_blank)
		(*t) |= BLANK_MASK;
	return TRUE;
}


static void
get_letter (GtkWidget *widget, gpointer data)
{
	Tile t = gtk_entry_get_text(GTK_ENTRY(text))[0];
	gtk_widget_destroy(GTK_WIDGET(dialog));

	if ( !is_valid_tile( &t ) )
	  {
		get_blank();
	  }
	else
	  {
		int x_ret = board_position.x + mover.xtarget*tile_size.x;
		int y_ret = board_position.y + mover.ytarget*tile_size.y;

		t |= BLANK_MASK;
		board_tile_state[mover.xtarget][mover.ytarget].t = t;

		draw_tile (buffer,NULL,x_ret,y_ret,t);
  		
		gdk_draw_pixmap (space->window,
				space->style->fg_gc[GTK_WIDGET_STATE(space)],
				buffer, x_ret-1, y_ret-1,
				x_ret-1, y_ret-1,
				tile_size.x+2, tile_size.y+2);
	  }
	
}

static void
get_blank ()
{
	GtkWidget *okay_button;

	dialog = gtk_dialog_new ();	
	gtk_window_set_policy (GTK_WINDOW(dialog), FALSE, FALSE, TRUE);
	gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);
	gtk_window_set_title (GTK_WINDOW(dialog), "Enter Letter:");
	okay_button = gtk_button_new_with_label("Okay");
	text = gtk_entry_new_with_max_length (1);

	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
								okay_button);
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
								text);	

	gtk_signal_connect (GTK_OBJECT (okay_button), "clicked",
						GTK_SIGNAL_FUNC (get_letter), NULL);

	
	gtk_widget_show_all (dialog);

}


static void
invitation_response ( GtkWidget *widget, gpointer data )
{
	gtk_widget_destroy(GTK_WIDGET(dialog));
	if ( *((char *)data) == 'A' )
	  {
		int retVal = pci_accept_invite(TRUE);
		if( retVal == SUCCESS )
		  {
		  	print_msg("> Invitation was successfully accepted\n");
		  }
		else
		  {
		  	print_msg( status_messages[retVal] );
		  }
	  }
	else
	  {
		pci_accept_invite(FALSE);
	  }
}



void
get_invitation_response (PlayerID pid)
{
	char label[45];

	GtkWidget *accept;
	GtkWidget *deny;
	GtkWidget *prompt;
	GtkWidget *hbox;

	sprintf( label, "You have been invited\nto play by: %i", pid );

	prompt = gtk_label_new( label );

	dialog = gtk_dialog_new ();	
	gtk_window_set_policy (GTK_WINDOW(dialog), FALSE, FALSE, TRUE);
	gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);

	accept = gtk_button_new_with_label("Accept");
	deny = gtk_button_new_with_label("Deny");

	hbox = gtk_hbox_new (FALSE, 0);	
	gtk_box_pack_start (GTK_BOX (hbox), accept, TRUE, TRUE, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox), deny, TRUE, TRUE, FALSE);

	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
								hbox);
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
								prompt);	

	gtk_signal_connect (GTK_OBJECT (accept), "clicked",
						GTK_SIGNAL_FUNC (invitation_response), "A");

	gtk_signal_connect (GTK_OBJECT (deny), "clicked",
						GTK_SIGNAL_FUNC (invitation_response), "D");

	
	gtk_widget_show_all (dialog);

}


static void
setup_rack ()
{
	int i;
	
	rack_size = default_rack_size;
	rack_state = s_malloc (rack_size * sizeof (Tile));
	
	for (i=0; i<rack_size; i++)
	  {
		rack_state[i] = TILE_EMPTY;
	  }

}



static void
setup_board ()
{
	int i;
	
	/* This sets up the board dimensions
	   If the board is allowed to change dimensions, this can be changed
	   to accomodate a message transfer with the game manager to get the
	   dimensions at runtime
	*/
	board_dim.x = default_board_dim_x;
	board_dim.y = default_board_dim_y;

	/* Setting up the matrix in memory for the board, and the tile info */
	board_square_state = s_malloc (board_dim.x * sizeof (Board_Square *));
	board_tile_state = s_malloc (board_dim.x * sizeof (Location_Status *));

	for (i=0; i<board_dim.x; i++)
	  {
		board_square_state[i] = s_malloc (board_dim.y * sizeof (Board_Square));
		board_tile_state[i] = s_malloc (board_dim.y * sizeof (Location_Status));
	  }

	clear_board ();
	set_board_squares ();	
}

static void
setup_move ()
{
	board_move.pos = s_malloc (rack_size*sizeof(Coord));
	board_move.t = s_malloc (rack_size*sizeof(Tile));
	board_move.is_swap = FALSE;
}


static void
clear_board ()
{
	int i, j;
	for (i=0; i<board_dim.x; i++)
		for (j=0; j<board_dim.y; j++)
		  {
			board_tile_state[i][j].t = TILE_EMPTY;
			board_tile_state[i][j].new_m = FALSE;
			board_tile_state[i][j].fixed = FALSE;						
		  }
	
}


static void
set_board_squares ()
{
	int i, j;
	
	/* This is just a hardcoded setup of the board
	   If the board changes, this can easily be accomodated to a
	   send grid location - receive text and colour scheme */

	for (i=0; i<board_dim.x; i++)
		for (j=0; j<board_dim.y; j++)
		  { 
			board_square_state[i][j].color = default_board_stats[i][j];
			board_square_state[i][j].label =
						board_text[default_board_stats[i][j]];
		  }
}




static void
free_mem ()
{
	int i;
	for(i=0; i<board_dim.x; i++)
	  {
	  	s_free (board_square_state[i]);
		s_free (board_tile_state[i]);
	  }
	s_free (board_square_state);
	s_free (board_tile_state);

	s_free (rack_state);

	s_free (board_move.pos);
	s_free (board_move.t);
}




static void
create_window()
{
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(GTK_WINDOW(window), APP_DEFAULT_WIDTH, APP_DEFAULT_HEIGHT);
	gtk_window_set_title (GTK_WINDOW(window), APPNAME);
	gtk_signal_connect (GTK_OBJECT (window), "delete_event",
						GTK_SIGNAL_FUNC (menu_game_end), NULL);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (window), vbox);
	gtk_widget_show (vbox);

}
	



static void
create_space()
{
	GtkWidget *b_TryMove;
	GtkWidget *b_Challenge;
	GtkWidget *b_MakeMove;
	GtkWidget *b_Pass;	

	GtkWidget *vscroll;
	GtkWidget *hbox;

	table = gtk_table_new (4, 3, FALSE);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, FALSE);
	gtk_widget_show (table);

	space = gtk_drawing_area_new ();
	gtk_table_attach (GTK_TABLE (table), space, 0, 1, 0, 4,
					GTK_FILL | GTK_SHRINK | GTK_EXPAND,
					GTK_FILL | GTK_SHRINK | GTK_EXPAND,
					0, 0);
	gtk_widget_show (space);

	gtk_widget_set_events (space, GDK_EXPOSURE_MASK |
							GDK_BUTTON_PRESS_MASK |
							GDK_POINTER_MOTION_MASK |
							GDK_BUTTON_RELEASE_MASK);
		

	gtk_signal_connect(GTK_OBJECT(space), "expose_event",
						GTK_SIGNAL_FUNC(expose_space), NULL);
	gtk_signal_connect(GTK_OBJECT(space), "configure_event",
						GTK_SIGNAL_FUNC(configure_space), NULL);
	gtk_signal_connect(GTK_OBJECT(space), "button_press_event",
						GTK_SIGNAL_FUNC(button_press_space), NULL);
	gtk_signal_connect(GTK_OBJECT(space), "button_release_event",
						GTK_SIGNAL_FUNC(button_release_space), NULL);
	gtk_signal_connect(GTK_OBJECT(space), "motion_notify_event",
						GTK_SIGNAL_FUNC(motion_notify_space), NULL);

	b_TryMove = gtk_button_new_with_label ("Try Move");
	b_Pass = gtk_button_new_with_label ("Swap Tiles");
	b_MakeMove = gtk_button_new_with_label ("Make Move");
	b_Challenge = gtk_button_new_with_label ("Challenge");

	gtk_signal_connect (GTK_OBJECT (b_TryMove), "clicked",
						GTK_SIGNAL_FUNC (button_try_move), NULL);
	gtk_signal_connect (GTK_OBJECT (b_Pass), "clicked",
						GTK_SIGNAL_FUNC(button_pass), NULL);
	gtk_signal_connect (GTK_OBJECT (b_MakeMove), "clicked",
						GTK_SIGNAL_FUNC(button_make_move), NULL);
	gtk_signal_connect (GTK_OBJECT (b_Challenge), "clicked",
						GTK_SIGNAL_FUNC(button_challenge), NULL);


	gtk_table_attach (GTK_TABLE (table), b_TryMove, 1, 2, 2, 3,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK,
					2, 2);
	gtk_table_attach (GTK_TABLE (table), b_Pass, 2, 3, 2, 3,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK,
					2, 2);
	gtk_table_attach (GTK_TABLE (table), b_MakeMove, 1, 2, 3, 4,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK,
					2, 2);
	gtk_table_attach (GTK_TABLE (table), b_Challenge, 2, 3, 3, 4,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK,
					2, 2);

	gtk_widget_show (b_TryMove);
	gtk_widget_show (b_Pass);
	gtk_widget_show (b_MakeMove);
	gtk_widget_show (b_Challenge);
	
	
	t_Scores = gtk_text_new (NULL, NULL);
	t_Messages = gtk_text_new (NULL, NULL);
	gtk_text_set_editable ( GTK_TEXT (t_Scores), FALSE);
	gtk_text_set_editable ( GTK_TEXT (t_Messages), FALSE);
	gtk_text_set_word_wrap ( GTK_TEXT (t_Scores), TRUE);
	gtk_text_set_word_wrap ( GTK_TEXT (t_Messages), TRUE);


	hbox = gtk_hbox_new (FALSE, 0);	
	gtk_table_attach (GTK_TABLE (table), hbox, 1, 3, 1, 2,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK | GTK_EXPAND,
					2, 2);
	gtk_widget_show (hbox);

	gtk_box_pack_start (GTK_BOX (hbox), t_Messages, TRUE, TRUE, FALSE);
	gtk_widget_show (t_Messages);
	
	vscroll = gtk_vscrollbar_new (GTK_TEXT (t_Messages)->vadj);
	gtk_box_pack_end (GTK_BOX (hbox), vscroll, FALSE, 0, 0);
	gtk_widget_show (vscroll);
	
	gtk_table_attach (GTK_TABLE (table), t_Scores, 1, 3, 0, 1,
					GTK_FILL | GTK_SHRINK,
					GTK_FILL | GTK_SHRINK,
					2, 2);
	gtk_widget_show (t_Scores);
	
	
	gtk_widget_set_usize (space, 500, 500);
	gtk_widget_set_usize (hbox, 300, 285);
	gtk_widget_set_usize (t_Scores, 300, 55);

}


static void
create_buffer (GtkWidget *widget)
{
	if (buffer)
	  {
		gdk_pixmap_unref(buffer);
	  }

	buffer = gdk_pixmap_new(widget->window,
		widget->allocation.width, widget->allocation.height, -1);
}



void
print_msg (char *msg)
{
	gtk_text_insert ( GTK_TEXT(t_Messages), NULL, NULL, NULL, msg, -1 );
}


void
print_scores (Scores *s)
{
	int i;
	char temp[50];

	gtk_text_set_point ( GTK_TEXT(t_Scores), 0 );

	for (i=0; i<s->num_players; i++)
	  {
	    if( s->players[i] == myid )
			sprintf(temp, "You                     %i\n", s->scores[i]);
		else if( s->players[i] < 0 )
			sprintf(temp, "Auto Player          %i\n", s->scores[i]);
		else
		  	sprintf(temp, "%i                         %i\n", s->players[i], s->scores[i]);
		gtk_text_insert ( GTK_TEXT(t_Scores), NULL, NULL, NULL, temp, -1);
	  }
	  
	gtk_text_insert ( GTK_TEXT(t_Scores), NULL, NULL, NULL, "\n\n\n\n\n\n\n", -1);
	gtk_text_set_point ( GTK_TEXT(t_Scores), 0 );
	gtk_text_insert (GTK_TEXT(t_Scores), NULL, NULL, NULL, "PlayerID           Score\n", -1);

}


static void
refresh (GtkWidget *widget)
{
	Coord tile_size_temp;
	
	tile_size_temp.x = (widget->allocation.width - 2*default_border_width) /
					(board_dim.x);
	tile_size_temp.y = (widget->allocation.height / board_aspect_ratio) /
					(board_dim.y + 2);
	
	tile_size.x = (tile_size_temp.x > tile_size_temp.y)?
					tile_size_temp.y : tile_size_temp.x;
	tile_size.y = tile_size.x * board_aspect_ratio;

	board_position.x = (widget->allocation.width -
							(tile_size.x * board_dim.x)) / 2;
	board_position.y = (widget->allocation.height -
					(tile_size.y * (board_dim.y+2))) / 2;
	if (board_position.y < default_border_width)
		board_position.y = default_border_width;


	rack_position.x = board_position.x + (tile_size.x * (board_dim.x -
						rack_size) / 2);
	rack_position.y = board_position.y + (tile_size.y * (board_dim.y+0.5));

	create_mover ();
	redraw_board ();		
}



static gint
expose_space (GtkWidget *widget, GdkEventExpose *event)
{
	gdk_draw_pixmap (widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
			buffer, event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

	return (FALSE);
}


static gint
configure_space (GtkWidget *widget, GdkEventConfigure *event)
{
	create_buffer (widget);
	refresh (widget);
	return (TRUE);
}



static gint
button_press_space (GtkWidget *widget, GdkEventButton *event)
{
	if (event->button == 1)
	  {
		if(board_move.is_swap)
		  {
		  	swap_tile(event->x, event->y);
		  }
	  	else
		  {
		  	if (button_down == 1)
			  {
				setup_mover (event->x, event->y, RELEASE);
				button_down = 0;
				return FALSE;
			  }
			if (setup_mover (event->x, event->y, PRESS))
			  {
			  	button_down = 1;
			  }
		  }
	  }
	return (FALSE);
}


static gint
button_release_space (GtkWidget *widget, GdkEventButton *event)
{
	if (event->button == 1)
	  {
	  	if (button_down == 1)
		  {
		    setup_mover (event->x, event->y, RELEASE);
		  }
		button_down = 0;
	  }
	return(FALSE);
}


static gint
motion_notify_space (GtkWidget *widget, GdkEventButton *event)
{
	if (button_down == 1)
	  {
	  	gdk_window_move( mover.window, event->x-mover.xoff,
						event->y-mover.yoff);
	  }
	return(FALSE);
}



